/*
*   slow: print slowly.
*   92/07/28 John H. DuBois III   john@armory.com
*   92/09/13 Converted to not use stdio, 
*            and to read chars in same units as they are written.
*   94/01/29 Added h & r options.
*   94/01/31 Added b option, worked around itimers bugs.
*   94/02/06 Added l option.
*   94/03/28 Ported to XENIX.  If compiled without M_UNIX defined,
*            will use nap() as an inferior substitute for itimers;
*            also avoids use of sigaction.
*   94/12/23 Expanded help info, added s option.
*   95/05/24 Allow k or m suffix on command line values.
*   96/07/03 Write help to fd 1, not 2.
*/
char *Comment = "@(#) slow 1.1 96/07/03";

#include <sys/types.h>
#include <sys/timeb.h>
#include <sys/time.h>
#ifdef M_UNIX
#   include <signal.h>
#endif
#ifdef DEBUG
#   include <stdio.h>
#endif
#include <fcntl.h>
#include <memory.h>
#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

#define WriteString(S,fd) write(fd,S,strlen(S))
#define ErrExit(Message) WriteString(Message,2),exit(1)
#ifndef NULL
#   define NULL ((char *) 0)
#endif
#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) < (b)) ? (b) : (a))

#ifndef M_UNIX
struct sigaction {
    int bogus;
};
extern char *sys_errlist[];
#define strerror(errno) (sys_errlist[errno])
#endif

void AlarmCatch(int foo);

char *Buf;
unsigned BufferSize = 16384;

main(argc,argv)
int argc;
char *argv[];
{
    int bps;		/* Blocks per second to print */
    int InFile;		/* File being read */
    int i;		/* Argument being processed */
    int NapTime;	/* Time to sleep between writes, in uS */
    /* Number of nanoseconds to sleep between writing a block (average!) */
    int nSPerBlock;
    int Repeat = 1;	/* Number of times to repeat each file */
    int BS = 1;		/* Block size */
    struct sigaction Ignore,Catch;
#ifdef M_UNIX
    struct itimerval SleepTime;	/* Time between wakeups */
#endif
    int Count;		/* Number of times file has been output */
    /* Max amount of time we will try to catch up on output for, in nS.  
       If we oversleep for an amount of time larger than this, it's 
       capped at MaxLag. */
    int MaxLag = 100;
    char *HzVar;	/* HZ environment var value */
    int Err = 0;
    char c;			/* Option character */
    extern char *optarg;	/* For optargs */
    extern int optind;		/* For optargs */
    char Usage[] = 
    "Usage: slow [-h] [-r<repeat>] [-b<blocksize>] [-l<lagtime>]\n"\
    "            [-s<buffer size>] <bps> [file ...]\n";

    while ((c = getopt(argc,argv,"hr:b:l:s:")) != -1)
	switch (c) {
	case 'h':
	    WriteStrings(1,
"slow: Write data slowly.\n"\
"     At regular intervals, slow reads a given number of characters from\n"\
"its input (files or its standard input) and writes them to its standard\n"\
"output.  The number of characters read and written is referred to as a\n"\
"block, and by default is one character.  The interval is the reciprocal of\n"\
"the blocks-per-second argument that is slow's one required argument.\n"\
"     If slow determines that more than one interval has passed without\n"\
"characters being written, it will process more than one block to catch up.\n"\
"This will occur under various circumstances.  If the interval is less than\n"\
"the shortest amount of time that a process can sleep, slow will\n"\
"occasionally, or always, have to process more than one block.  slow may\n"\
"also oversleep due to CPU load or being suspended, or due to being blocked\n"\
"on input or output as a result of IO throttling, data available at an\n"\
"insufficient rate, flow control, etc. See the -l argument to control this\n"\
"behaviour.\n"\
"     The largest amount of data that slow will write is one buffer's\n"\
"worth.  If more data is to be written in an operation, either because the\n"\
"blocksize is larger than the buffer size or because enough blocks are\n"\
"being written that the buffer size is exceeded, data is output in\n"\
"consecutive buffer-size writes until the required amount has been written.\n"\
"The last write may be less than a buffer's worth.  If it is important that\n"\
"each write have a specific size, set both the buffer size and block size\n"\
"to that value.\n",
Usage,
"where <bps> is the number of blocks per second to print.\n"\
"If no filenames are given, the standard input is read.\n"\
"Options:\n"\
"-h: Print this help.\n"\
"The values given with any of the following options can include a k or m\n"\
"suffix to cause the value to be multiplied by 1K (1024) or 1M (1048576).\n"\
"-b<blocksize>: Set the block size to <blocksize> characters.\n"\
"    If 0 is given, the block size become the size of the entire file.\n"\
"-s<buffer size>: Set the buffer size.  The default is 16384 bytes.\n"\
"-l<lagtime>: The maximum amount of time that slow will try to catch up on,\n"\
"    in milliseconds.  If slow oversleeps more than <lagtime>, only\n"\
"    <lagtime>'s worth of blocks is output.  The default is 100 mS.  Use 0\n"\
"    to eliminate the cap.  Data is never discarded, only written later.\n"\
"-r<repeat>: Output each file <repeat> times.  If <repeat> is 0, the first\n"\
"    file named is repeated indefinitely.  The program may be halted by\n"\
"    keyboard interrupt.\n",
(char *)0);
	    exit(0);
	case 'r':
	    if ((Repeat = katoi(optarg)) == -1) {
		WriteString("slow: invalid repeat value.\n",2);
		Err = 1;
	    }
	    break;
	case 'l':
	    if ((MaxLag = katoi(optarg)) == -1 || MaxLag > 1000) {
		WriteString("slow: invalid lagtime value.\n",2);
		Err = 1;
	    }
	    break;
	case 'b':
	    if ((BS = katoi(optarg)) == -1) {
		WriteString("slow: invalid block size.\n",2);
		Err = 1;
	    }
	    /* Intead of actually setting BS to the file size (which
	       will likely be different between files), just make it
	       as big as a file can be */
	    if (!BS)
		BS = ( (((unsigned)1)<<31))-1;
	    break;
	case 's':
	    if ((BufferSize = katoi(optarg)) < 1) {
		WriteString("slow: invalid buffer size.\n",2);
		Err = 1;
	    }
	    break;
	case '?':
	    Err = 1;
	}
    if ((argc - optind) < 1) {
	WriteString("slow: no bps value given.\n",2);
	Err = 1;
    }
    else if ((bps = katoi(argv[optind++])) < 1 || bps > 100000000) {
	WriteString("slow: invalid bps value.\n",2);
	Err = 1;
    }
    if (Err) {
	WriteString(Usage,2);
	exit(1);
    }
    
#ifdef DEBUG
    printf("Repeat=%d MaxLag=%d BlockSize=%d BufferSize=%d BPS=%d\n",
    Repeat,MaxLag,BS,BufferSize,bps);
#endif
    if (!(Buf = malloc(BufferSize)))
	ErrExit("Could not allocate a buffer.\n");
    NapTime = 1000000 / bps;
    if (!NapTime)
	NapTime = 1;
    /* itimers bug makes signal never occur if usec part is zero */
    if (!(NapTime % 1000000))
	NapTime++;
    nSPerBlock = 1000000000 / bps;

    /* Convert to nS and add on the amount of time that we *should* sleep;
       account for it here to simplify later calculations */
    MaxLag = MaxLag * 1000000 + (HzVar = getenv("HZ")) ?
    max(1000000000/atoi(HzVar),NapTime) : 10000000;

#ifdef M_UNIX
    /* Set alarm interval */
    /* itimers ignore usec > 1M, so split it up */
    SleepTime.it_interval.tv_sec = NapTime / 1000000;
    SleepTime.it_interval.tv_usec = NapTime % 1000000;
    SleepTime.it_value.tv_sec = NapTime / 10000000;
    SleepTime.it_value.tv_usec = NapTime % 1000000;
    
    /* Set up actions for blocking & allowing alarm signal */
    Ignore.sa_handler = SIG_IGN;
    sigemptyset(&Ignore.sa_mask);
    Ignore.sa_flags = 0;
    Catch = Ignore;
    Catch.sa_handler = AlarmCatch;

    sigaction(SIGALRM, &Ignore, NULL);
    setitimer(ITIMER_REAL,&SleepTime,NULL);	/* Begin alarms */
#endif

    if ((argc - optind) < 1) {
	if (Repeat != 1) {
	    WriteString("slow: Can't use a repeat value on stdin.\n",2);
	    exit(1);
	}
	SlowPrint(0,nSPerBlock,NapTime,&Ignore,&Catch,BS,MaxLag);
    }
    else
	for (i = optind; i < argc; i++)
	    if ((InFile = open(argv[i],O_RDONLY)) == -1)
		WriteStrings(2,"slow: could not open file \"",
		argv[i],"\": ",strerror(errno),"\n",(char *) 0);
	    else {
		/* If Repeat is 0, do forever */
		for (Count = 0; !Repeat || Count < Repeat; Count++) {
		    lseek(InFile,0L,SEEK_SET);
		    SlowPrint(InFile,nSPerBlock,NapTime,&Ignore,&Catch,BS,
		    MaxLag);
		}
		close(InFile);
	    }
}

/* Convert string *s to a non-negative integer.  Return -1 if s does not
   contain a number, if it contains a negative number, or if it includes
   any characters that are not part of the number.   A k or m suffix
   causes the value to be multipled by 1k or 1m. */
katoi(char *s)
{
    int i;
    char *ptr;

    i = strtol(s,&ptr,0);
    if (ptr == s || i < 0)
	return -1;
    switch (*ptr) {
    case 'k':
    case 'K':
	i <<= 10;
	break;
    case 'm':
    case 'M':
	i <<= 20;
	break;
    case '\0':
	break;
    default:
	return -1;
    }
    return i;
}

SlowPrint(File,nSPerBlock,NapTime,Ignore,Catch,BS,MaxLag)
int File;
int nSPerBlock;
int NapTime;
int MaxLag;
struct sigaction *Ignore,*Catch;
int BS;
{
    int nSNapped;	/* Time we actually napped, in nS */
    int BlocksToPrint;	/* Number of blocks to print this time */
    int CharsToPrint;	/* Number of chars to print this time */
    /* Number of nS slept beyond those that will be accounted for by
       the blocks we write this time (fractional-block-time) */
    int LeftoverNapTime;
    struct timeb Time,LastTime;	/* How long we actually slept */
    int NumRead = 1;	/* How many chars were read */

    ftime(&LastTime);				/* Initialize LastTime */
    while (NumRead > 0) {
#ifdef M_UNIX
	sigaction(SIGALRM, Catch, NULL);
	pause();
	sigaction(SIGALRM, Ignore, NULL);
#else
	nap(NapTime/1000);
#endif

	/* Deal with ftime bugs by using only millitm */
	ftime(&Time);
	nSNapped = (Time.millitm - LastTime.millitm) * 1000000;
#ifdef BADFTIME
	if (nSNapped < 0)
#else
	if (Time.time != LastTime.time)
#endif
	    nSNapped += 1000000000;

	/* Put a cap on lag time */
	if (MaxLag && nSNapped > MaxLag)
	    nSNapped = MaxLag;
	LastTime = Time;
	/* Print a number of blocked based on the leftover fractional block
	time from last iteration, and the amount of time we last slept */
	BlocksToPrint = (nSNapped + LeftoverNapTime) / nSPerBlock;
	if (BlocksToPrint < 1)
	    BlocksToPrint = 1;
	LeftoverNapTime = nSNapped - BlocksToPrint * nSPerBlock;
	CharsToPrint = BlocksToPrint * BS;

	while (CharsToPrint && NumRead > 0) {
	    if ((NumRead = read(File,Buf,min(CharsToPrint,BufferSize))) > 0)
		write(1,Buf,NumRead);
	    CharsToPrint -= NumRead;
	}
    }
}

void AlarmCatch(int foo)
{
    ;
}

#ifdef __ICC
WriteStrings(fd)
#else
WriteStrings(int fd, ...)
#endif
{
    va_list ap;
    char *S;

    va_start(ap,fd);
    while (S = va_arg(ap,char *))
	write(fd,S,strlen(S));
    va_end(ap);
}
